import logging
import re
from threading import local
from typing import Iterable, Optional, NamedTuple, Union

_LOGGER = logging.getLogger(__name__)
_RE_BASIC_ACK = re.compile('^basic.ack$', re.IGNORECASE)

try:
    import pika
    _BasicProperties = pika.BasicProperties
except ImportError as e:
    # This module can not work without `pika` installed, but at least
    # we can allow Sphinx to successfully import the module.
    _LOGGER.error(e)
    _BasicProperties = object


class MessageProperties(_BasicProperties):
    """Basic message properties

    This is an alias for :class:`pika.BasicProperties`.
    """


class Message(NamedTuple):
    """A `typing.NamedTuple` representing a RabbitMQ message to be send

    :param exchange: RabbitMQ exchange name
    :param routing_key: RabbitMQ routing key
    :param body: The message's body
    :param properties: Message properties (see :class:`pika.BasicProperties`)
    """

    exchange: str
    routing_key: str
    body: Union[str, bytes]
    properties: MessageProperties


class DeliverySet:
    def __init__(self, start_tag, message_count):
        self.start_tag = start_tag
        self.confirmed_array = [False for _ in range(start_tag, start_tag + message_count)]
        self.confirmed_count = 0
        self.min_unconfirmed_tag = start_tag

    def confirm(self, delivery_tag, multiple=False):
        confirmed_array = self.confirmed_array
        confirmed_count = 0
        start_tag = self.start_tag
        idx = delivery_tag - start_tag
        if multiple:
            for i in range(self.min_unconfirmed_tag - start_tag, idx + 1):
                if i >= 0 and confirmed_array[i] is False:
                    confirmed_array[i] = True
                    confirmed_count += 1
            next_tag = delivery_tag + 1
            if self.min_unconfirmed_tag < next_tag:
                self.min_unconfirmed_tag = next_tag
        else:
            if idx >= 0 and confirmed_array[idx] is False:
                confirmed_array[idx] = True
                confirmed_count += 1

        self.confirmed_count += confirmed_count
        return confirmed_count > 0

    @property
    def all_confirmed(self):
        return self.confirmed_count == len(self.confirmed_array)


class DeliveryError(Exception):
    """A failed attempt to deliver messages."""


class ConnectionError(DeliveryError):
    """Can not connect to the server."""


class TimeoutError(DeliveryError):
    """The attempt to deliver messages has timed out."""


class Publisher:
    """A RabbitMQ message publisher

    Each instance maintains a separate RabbitMQ connection in every
    thread. If a connection has not been used for longer than the
    heartbeat interval set for the connection, it will be
    automatically closed. A new connection will be open when needed.

    :param app: Optional Flask app object. If not provided `init_app`
      must be called later, providing the Flask app object.
    :param url_config_key: Optional configuration key for the
      RabbitMQ's connection URL

    Example::

        from flask import Flask
        from flask_sqlalchemy import SQLAlchemy
        from flask_signalbus import rabbitmq

        app = Flask(__name__)

        headers = {'header1': 'value1', 'header2': 'value2'}
        properties = rabbitmq.MessageProperties(
            app_id='example-publisher',
            content_type='application/json',
            headers=headers,
        )
        m1 = rabbitmq.Message(exchange='', routing_key='test',
                              body='Message 1', properties=properties)
        m2 = rabbitmq.Message(exchange='', routing_key='test',
                              body='Message 2', properties=properties)

        mq = rabbitmq.Publisher(app)
        mq.publish_messages([m1, m2])

    """

    def __init__(self, app=None, *, url_config_key='SIGNALBUS_RABBITMQ_URL'):
        self._state = local()
        self._url_config_key = url_config_key
        if app is not None:
            self.init_app(app)

    def init_app(self, app):
        """Bind the instance to a Flask app object.

        :param app: A Flask app object
        """
        self.app = app
        self._url = app.config[self._url_config_key]
        self._kill_connection()

    @property
    def _channel(self):
        """The channel for the current thread."""
        return getattr(self._state, 'channel', None)

    @_channel.setter
    def _channel(self, new_channel):
        state = self._state
        old_channel = getattr(state, 'channel', None)
        if new_channel is not old_channel:
            if not (old_channel is None or old_channel.is_closed or old_channel.is_closing):
                old_channel.close()
            state.channel = new_channel
            state.message_number = 0

    def _get_connection(self):
        """This method returns the RabbitMQ connection for the current thread.
        If there are no connection, a new one will be created.
        """
        connection = getattr(self._state, 'connection', None)
        if connection is None:
            _LOGGER.info('Connecting to %s', self._url)

            # Tries to connect to RabbitMQ, returning the connection
            # handle. When the connection is established, the
            # _on_connection_open method will be invoked by pika.
            connection = self._state.connection = pika.SelectConnection(
                pika.URLParameters(self._url),
                on_open_callback=self._on_connection_open,
                on_open_error_callback=self._on_connection_open_error,
                on_close_callback=self._on_connection_closed,
            )
        return connection

    def _kill_connection(self):
        """This method closes the RabbitMQ connection for the current thread,
        and throws away the connection object. If there is a channel
        corresponding to the connection, it will be closed and throw
        away too.
        """
        self._channel = None
        connection = getattr(self._state, 'connection', None)
        if connection is not None:
            if not (connection.is_closed or connection.is_closing):
                connection.close()
            self._state.connection = None

    def _on_connection_open(self, connection):
        """This method is called by pika once the connection to RabbitMQ has
        been established. It passes the handle to the connection
        object.
        """
        _LOGGER.info('Connection opened')
        _LOGGER.info('Creating a new channel')

        # This will open a new channel with RabbitMQ by issuing the
        # Channel.Open RPC command. When RabbitMQ confirms the channel
        # is open by sending the Channel.OpenOK RPC reply, the
        # _on_channel_open method will be invoked.
        connection.channel(on_open_callback=self._on_channel_open)

    def _on_connection_open_error(self, troubled_connection, err):
        """This method is called by pika if the connection to RabbitMQ can not
        be established. In this case, we will throw away the
        connection object.
        """
        _LOGGER.error('Connection open failed: %s', err)
        troubled_connection.ioloop.stop()
        if getattr(self._state, 'connection', None) is troubled_connection:
            self._kill_connection()
        self._state.error = ConnectionError(err)

    def _on_connection_closed(self, closed_connection, reason):
        """This method is invoked by pika when the connection to RabbitMQ is
        closed unexpectedly. In this case, we will throw away the
        connection object.
        """
        _LOGGER.info('Connection closed: %s', reason)
        closed_connection.ioloop.stop()
        if getattr(self._state, 'connection', None) is closed_connection:
            self._kill_connection()

    def _on_channel_open(self, channel):
        """This method is invoked by pika when the channel has been opened.
        We use the passed channel object to immediately send the
        queued messages (asynchronously, no blocking).
        """
        _LOGGER.info('Channel opened')
        self._channel = channel
        channel.add_on_close_callback(self._on_channel_closed)

        # Send the Confirm.Select RPC method to RabbitMQ to enable
        # delivery confirmations on the channel. The only way to turn
        # this off is to close the channel and create a new one.
        # When the message is confirmed from RabbitMQ, the
        # _on_delivery_confirmation method will be invoked passing in a
        # Basic.Ack or Basic.Nack method from RabbitMQ that will
        # indicate which messages it is confirming or rejecting.
        _LOGGER.info('Issuing Confirm.Select RPC command')
        channel.confirm_delivery(self._on_delivery_confirmation)

        assert channel.is_open
        self._send_messages(channel)

    def _on_channel_closed(self, closed_channel, reason):
        """This method is invoked by pika when RabbitMQ unexpectedly closes
        the channel.  Channels are usually closed if you attempt to do
        something that violates the protocol, such as re-declare an
        exchange or queue with different parameters. In this case, we
        will close the connection and throw away the connection
        object.
        """
        _LOGGER.info('Channel %i was closed: %s', closed_channel, reason)
        if self._channel is closed_channel:
            self._kill_connection()

    def _on_delivery_confirmation(self, method_frame):
        """This method is invoked by pika when RabbitMQ responds to a
        Basic.Publish RPC command, passing in either a Basic.Ack or
        Basic.Nack frame with the delivery tag of the message that was
        published. The delivery tag is an integer counter indicating
        the message number that was sent on the channel via
        Basic.Publish.

        NOTE: When possible, RabbitMQ sends one delivery confirmation
        for multiple messages. In this case
        `method_frame.method.multiple` will be `True`, meaning that
        this delivery confirmation applies to all messages whose tag
        is less or equal than `method_frame.method.delivery_tag`.
        """
        state = self._state
        connection = getattr(state, 'connection', None)
        if connection is None:
            _LOGGER.warning(
                'A message delivery confirmation will be ignored because a connection '
                'object is not available. This should happen very rarely, or never.')
            return

        method = method_frame.method
        medhod_name = method.NAME
        tag = method.delivery_tag
        multiple = method.multiple
        _LOGGER.debug('Received %s for delivery tag: %i (multiple: %s)',
                      medhod_name, tag, multiple)

        if not _RE_BASIC_ACK.match(medhod_name):
            state.received_nack = True

        pending = state.pending_deliveries
        pending.confirm(tag, multiple)
        if pending.all_confirmed:
            if state.received_nack:
                state.error = DeliveryError('received nack')
            else:
                state.error = None
            connection.ioloop.stop()

    def _send_messages(self, channel):
        """This method publishes messages to RabbitMQ, creating a set of
        pending unconfirmed deliveries. This set will be used to check
        for delivery confirmations in the _on_delivery_confirmations
        method.
        """
        state = self._state
        messages = state.messages

        n = len(messages)
        state.pending_deliveries = DeliverySet(state.message_number + 1, n)
        state.received_nack = False
        state.message_number += n
        for m in messages:
            channel.basic_publish(m.exchange, m.routing_key, m.body, m.properties)
        _LOGGER.debug('Published %i messages', len(messages))

    def publish_messages(
            self,
            messages: Iterable[Message],
            *,
            timeout: Optional[int] = None,
            allow_retry: bool = True,
    ):
        """Publishes messages, waiting for delivery confirmations.

        This method will block until a confirmation from the RabbitMQ
        broker has been received for each of the messages.

        :param messages: The messages to publish
        :param timeout: Optional timeout in seconds

        """
        message_list = messages if isinstance(messages, list) else list(messages)
        if len(message_list) == 0:
            return

        state = self._state
        state.messages = message_list
        state.error = TimeoutError()

        channel = self._channel
        if channel is not None and channel.is_open:
            self._send_messages(channel)

        connection = self._get_connection()
        if timeout is not None:
            connection.ioloop.call_later(timeout, connection.ioloop.stop)

        try:
            connection.ioloop.start()
        except KeyboardInterrupt:
            if not (connection.is_closed or connection.is_closing):
                connection.close()
            if not connection.is_closed:
                # Start the loop again until the connection is fully
                # closed. The loop should stop on its own.
                connection.ioloop.start()
            raise

        error = getattr(state, 'error', None)

        # If at the end of the ioloop the connection is closed, most
        # probably the connection has been closed by the server. In
        # this case, we should retry with a fresh connection, but only
        # once.
        if allow_retry and connection.is_closed and not isinstance(error, ConnectionError):
            _LOGGER.debug('Re-executing publish_messages()')
            return self.publish_messages(message_list, timeout=timeout, allow_retry=False)

        if error is not None:
            raise error



import flask
import pytest
from sqlalchemy.engine import Engine
from sqlalchemy import event
import flask_sqlalchemy as fsa
import flask_signalbus as fsb
from mock import Mock


@event.listens_for(Engine, "connect")
def set_sqlite_pragma(dbapi_connection, connection_record):
    cursor = dbapi_connection.cursor()
    cursor.execute("PRAGMA foreign_keys=ON")
    cursor.close()


class SignalBusAlchemy(fsb.SignalBusMixin, fsa.SQLAlchemy):
    pass


class AtomicSQLAlchemy(fsb.AtomicProceduresMixin, fsa.SQLAlchemy):
    pass


@pytest.fixture
def app(request):
    app = flask.Flask(request.module.__name__)
    app.testing = True
    app.config['SQLALCHEMY_DATABASE_URI'] = 'sqlite:///:memory:'
    app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = False
    app.config['SQLALCHEMY_RECORD_QUERIES'] = True
    app.config['SIGNALBUS_RABBITMQ_URL'] = '?heartbeat=5'
    return app


@pytest.fixture(params=['direct', 'mixin_bound', 'mixin_init_app'])
def db(app, request):
    if request.param == 'direct':
        db = fsa.SQLAlchemy(app)
        db.signalbus = fsb.SignalBus(db)
    elif request.param == 'mixin_bound':
        db = SignalBusAlchemy(app)
    elif request.param == 'mixin_init_app':
        db = SignalBusAlchemy()
        db.init_app(app)
        db.app = app

    assert hasattr(db, 'signalbus')
    assert db.get_app()
    return db


@pytest.fixture
def atomic_db(app):
    db = AtomicSQLAlchemy()
    db.init_app(app)
    db.app = app
    assert db.get_app()
    return db


@pytest.fixture
def signalbus(app, db):
    return db.signalbus


@pytest.fixture
def signalbus_with_pending_signal(app, db, signalbus, Signal):
    signalbus.autoflush = False
    sig = Signal(name='signal', value='1')
    db.session.add(sig)
    db.session.commit()
    signalbus.autoflush = True
    return signalbus


@pytest.fixture
def signalbus_with_pending_error(app, db, signalbus, Signal):
    signalbus.autoflush = False
    sig = Signal(name='error', value='1')
    db.session.add(sig)
    db.session.commit()
    signalbus.autoflush = True
    return signalbus


@pytest.fixture
def send_mock():
    return Mock()


@pytest.fixture
def Signal(db, send_mock):
    class Signal(db.Model):
        __tablename__ = 'test_signal'
        id = db.Column(db.Integer, primary_key=True)
        name = db.Column(db.String(60))
        value = db.Column(db.String(60))

        def send_signalbus_message(self):
            properties = getattr(self, 'properties', [])
            assert hasattr(Signal, '__marshmallow__')
            dump = Signal.__marshmallow_schema__.dump(self)
            send_mock(
                self.id,
                self.name,
                self.value,
                {p.name: p.value for p in properties},
                dump.data if hasattr(dump, 'data') else dump,
            )
            if self.name == 'error':
                raise ValueError(self.value)

        signalbus_order_by = (id, db.desc(name))

    db.create_all()
    yield Signal
    db.drop_all()


@pytest.fixture
def SignalSendMany(db, send_mock):
    class SignalSendMany(db.Model):
        __tablename__ = 'test_signal_send_many'
        id = db.Column(db.Integer, primary_key=True)
        value = db.Column(db.String(60))

        signalbus_autoflush = False
        signalbus_burst_count = 100
        signalbus_order_by = (value,)

        def send_signalbus_message(self):
            pass

        @classmethod
        def send_signalbus_messages(cls, instances):
            for instance in instances:
                send_mock(instance.id)

    db.create_all()
    yield SignalSendMany
    db.drop_all()


@pytest.fixture
def SignalProperty(db, send_mock, Signal):
    class SignalProperty(db.Model):
        __tablename__ = 'test_signal_property'
        signal_id = db.Column(db.ForeignKey(Signal.id, ondelete='CASCADE'), primary_key=True)
        name = db.Column(db.String(60), primary_key=True)
        value = db.Column(db.String(60))
        signal = db.relationship(Signal, backref=db.backref("properties", passive_deletes='all'))

    db.create_all()
    yield SignalProperty
    db.drop_all()


@pytest.fixture
def NonSignal(db):
    class NonSignal(db.Model):
        __tablename__ = 'test_non_signal'
        id = db.Column(db.Integer, primary_key=True)

    db.create_all()
    yield NonSignal
    db.drop_all()


@pytest.fixture
def AtomicModel(atomic_db):
    db = atomic_db

    class AtomicModel(db.Model):
        __tablename__ = 'test_atomic_model'
        id = db.Column(db.Integer, primary_key=True)
        name = db.Column(db.String(60))
        value = db.Column(db.String(60))

    db.create_all()
    yield AtomicModel
    db.drop_all()



import logging
import time
from functools import wraps
from sqlalchemy.exc import DBAPIError

__all__ = [
    'DEADLOCK_ERROR_CODES',
    'DBSerializationError',
    'get_db_error_code',
    'retry_on_deadlock',
]


DEADLOCK_ERROR_CODES = ['40001', '40P01']


class DBSerializationError(Exception):
    """The transaction is rolled back due to a race condition."""


def get_db_error_code(exception):
    """Return 5-character SQLSTATE code, or '' if not available.

    Currently `psycopg2`, `psycopg2cffi`, and `MySQL Connector` are supported.
    """

    for attr in ['pgcode', 'sqlstate']:
        error_code = getattr(exception, attr, '')
        if error_code:
            break
    return error_code


def retry_on_deadlock(session, retries=7, min_wait=0.1, max_wait=10.0):
    """Return function decorator that executes the function again in case of a deadlock."""

    def decorator(action):
        """Function decorator that retries `action` in case of a deadlock."""

        @wraps(action)
        def f(*args, **kwargs):
            num_failures = 0
            while True:
                try:
                    return action(*args, **kwargs)
                except (DBAPIError, DBSerializationError) as e:
                    num_failures += 1
                    is_serialization_error = (
                        isinstance(e, DBSerializationError)
                        or get_db_error_code(e.orig) in DEADLOCK_ERROR_CODES
                    )
                    if num_failures > retries or not is_serialization_error:
                        raise
                session.rollback()
                if num_failures > 1:
                    wait_seconds = min(max_wait, min_wait * 2 ** (num_failures - 2))
                    time.sleep(wait_seconds)

        return f

    return decorator


def report_signal_count(signal_count):
    logger = logging.getLogger(__name__)
    if signal_count == 1:
        logger.info('%i signal has been successfully processed.', signal_count)
    elif signal_count > 1:
        logger.info('%i signals have been successfully processed.', signal_count)



import pytest
from flask_signalbus.utils import retry_on_deadlock, DBSerializationError
from sqlalchemy.exc import DBAPIError


def test_retry_on_deadlock(db):
    retry = retry_on_deadlock(db.session, retries=5, max_wait=0.0)
    executions = []

    @retry
    def f():
        executions.append(1)
        raise DBSerializationError

    @retry
    def g():
        executions.append(1)
        db.session.execute('xxx')

    with pytest.raises(DBSerializationError):
        f()
    assert len(executions) == 6

    with pytest.raises(DBAPIError):
        g()
    assert len(executions) == 7



import pytest
from mock import Mock
from sqlalchemy.orm import defer
from flask_signalbus.utils import DBSerializationError


def test_atomic(atomic_db):
    db = atomic_db
    commit = Mock()
    rollback = Mock()
    db.session.commit = commit
    db.session.rollback = rollback

    @db.atomic
    def f(x):
        return 1 / x

    with pytest.raises(ZeroDivisionError):
        f(0)
    commit.assert_not_called()
    rollback.assert_called_once()
    assert f(1) == 1
    commit.assert_called_once()
    rollback.assert_called_once()


def test_execute_atomic(atomic_db):
    db = atomic_db
    commit = Mock()
    rollback = Mock()
    db.session.commit = commit
    db.session.rollback = rollback
    var = 1

    with pytest.raises(RuntimeError):
        @db.execute_atomic
        def f1():
            raise RuntimeError
    commit.assert_not_called()
    rollback.assert_called_once()

    @db.execute_atomic
    def f2():
        assert var == 1
        return 666
    commit.assert_called_once()
    rollback.assert_called_once()
    assert f2 == 666

    assert db.execute_atomic(lambda: 777) == 777


def test_nested_execute_atomic(atomic_db):
    db = atomic_db
    commit = Mock()
    rollback = Mock()
    db.session.commit = commit
    db.session.rollback = rollback

    @db.execute_atomic
    def f1():
        @db.execute_atomic
        def f2():
            pass

    commit.assert_called_once()
    rollback.assert_not_called()


def test_retry_on_integrity_error(atomic_db, AtomicModel):
    db = atomic_db
    o = AtomicModel(
        id=1,
        name='test',
        value='1',
    )

    with pytest.raises(AssertionError):
        with db.retry_on_integrity_error():
            db.session.merge(o)
    assert len(AtomicModel.query.all()) == 0

    @db.execute_atomic
    def t1():
        with db.retry_on_integrity_error():
            return db.session.merge(o)
    assert len(AtomicModel.query.all()) == 1
    assert t1 not in db.session
    assert t1.name == 'test'
    assert t1.value == '1'

    db.session.expunge_all()
    o.value = '2'
    @db.execute_atomic
    def t2():
        with db.retry_on_integrity_error():
            db.session.merge(o)
    objects = AtomicModel.query.all()
    assert len(objects) == 1
    assert objects[0].value == '2'


# @pytest.mark.skip('too slow')
def test_retry_on_integrity_error_slow(atomic_db, AtomicModel):
    db = atomic_db
    call_list = []
    o = AtomicModel(
        id=1,
        name='test',
        value='1',
    )
    db.session.merge(o)
    db.session.commit()
    db.session.expunge_all()

    with pytest.raises(DBSerializationError):
        @db.execute_atomic
        def t():
            with db.retry_on_integrity_error():
                call_list.append(1)
                db.session.add(o)
    assert len(call_list) > 1


def test_get_instance(atomic_db, AtomicModel):
    db = atomic_db
    o = AtomicModel(id=1, name='test', value='1')
    assert o not in db.session
    assert AtomicModel.get_instance(o) is None
    db.session.add(o)
    assert AtomicModel.get_instance(o) is o
    assert o in db.session
    pk = o.id
    db.session.commit()
    assert AtomicModel.get_instance(pk) in db.session
    assert AtomicModel.get_instance((pk,)) in db.session
    assert AtomicModel.get_instance(o) in db.session


def test_get_instance_with_options(atomic_db, AtomicModel):
    db = atomic_db
    o = AtomicModel(id=1, name='test', value='1')
    assert o not in db.session
    assert AtomicModel.get_instance(o, defer('value')) is None
    db.session.add(o)
    assert AtomicModel.get_instance(o, defer('value')) is o
    assert o in db.session
    pk = o.id
    db.session.commit()
    assert AtomicModel.get_instance(pk, defer('value')) in db.session
    assert AtomicModel.get_instance((pk,), defer('value')) in db.session
    assert AtomicModel.get_instance(o, defer('value')) in db.session


def test_lock_instance(atomic_db, AtomicModel):
    db = atomic_db
    o = AtomicModel(id=1, name='test', value='1')
    assert o not in db.session
    assert AtomicModel.lock_instance(o) is None
    db.session.add(o)
    assert AtomicModel.lock_instance(o) is o
    assert o in db.session
    pk = o.id
    db.session.commit()
    assert AtomicModel.lock_instance(pk) in db.session
    assert AtomicModel.lock_instance((pk,)) in db.session
    assert AtomicModel.lock_instance(o) in db.session
    assert AtomicModel.lock_instance(o, read=True) in db.session


def test_get_pk_values(atomic_db, AtomicModel):
    o = AtomicModel(id=1, name='test', value='1')
    assert AtomicModel.get_pk_values(o) == (o.id,)
    assert AtomicModel.get_pk_values(1) == (1,)
    assert AtomicModel.get_pk_values((1,)) == (1,)



import threading
import signal
import sys

stopped = False


def stop_threads(*args, **kwargs):
    global stopped
    stopped = True


signal.signal(signal.SIGINT, stop_threads)
signal.signal(signal.SIGTERM, stop_threads)


def excepthook(args):
    if args.exc_type != SystemExit:
        print(f'Exception in thread {args.thread.name}:', file=sys.stderr)
        sys.excepthook(args.exc_type, args.exc_value, args.exc_traceback)
    stop_threads()


threading.excepthook = excepthook


class MyThread(threading.Thread):
    def __init__(self, throw=False):
        super().__init__()
        self.throw = throw

    def run(self):
        while not stopped:
            if self.throw:
                raise ValueError('Ops!')
        print(f'{self.name} exiting...')


threads = [MyThread() for _ in range(10)]
threads.append(MyThread(True))

for t in threads:
    t.start()

for t in threads:
    t.join()

exit_code = 1 if stopped else 0
sys.exit(exit_code)



"""
Adds to Flask-SQLAlchemy the capability to atomically send
messages (signals) over a message bus.
"""

import time
import logging
from sqlalchemy import event, inspect, and_, or_
from sqlalchemy.orm import mapper
from sqlalchemy.exc import DBAPIError
from flask_signalbus.utils import retry_on_deadlock, get_db_error_code, DEADLOCK_ERROR_CODES
from marshmallow_sqlalchemy import SQLAlchemyAutoSchema, ModelConversionError

__all__ = ['SignalBus', 'SignalBusMixin']


_SIGNALS_TO_FLUSH_SESSION_INFO_KEY = 'flask_signalbus__signals_to_flush'
_FLUSH_SIGNALS_LIMIT = 50000


def _get_class_registry(base):
    return base.registry._class_registry if hasattr(base, 'registry') else base._decl_class_registry


def _chunks(l, size):
    """Yield successive `size`-sized chunks from the list `l`."""

    for i in range(0, len(l), size):
        yield l[i:i + size]


def _raise_error_if_not_signal_model(model):
    if not hasattr(model, 'send_signalbus_message'):
        raise RuntimeError(
            '{} can not be flushed because it does not have a'
            ' "send_signalbus_message" method.'
        )


class SignalBusMixin(object):
    """A **mixin class** that can be used to extend
    :class:`~flask_sqlalchemy.SQLAlchemy` to handle signals.

    For example::

        from flask import Flask
        from flask_sqlalchemy import SQLAlchemy
        from flask_signalbus import SignalBusMixin

        class CustomSQLAlchemy(SignalBusMixin, SQLAlchemy):
            pass

        app = Flask(__name__)
        db = CustomSQLAlchemy(app)
        db.signalbus.flush()

    """

    def __init__(self, *args, **kwargs):
        super(SignalBusMixin, self).__init__(*args, **kwargs)
        event.listen(mapper, 'after_configured', _setup_schema(self.Model, self.session))

    def init_app(self, app, *args, **kwargs):
        """Bind the instance to a Flask app object.

        :param app: A Flask app object
        """

        super(SignalBusMixin, self).init_app(app, *args, **kwargs)
        self.signalbus._init_app(app)

    @property
    def signalbus(self):
        """The associated `SignalBus` object."""

        try:
            signalbus = self.__signalbus
        except AttributeError:
            signalbus = self.__signalbus = SignalBus(self, init_app=False)
        return signalbus


class SignalBus(object):
    """Instances of this class automatically send signal messages that
    have been recorded in the SQL database, over a message
    bus. Normally, the sending of the recorded messages (if there are
    any) is done after each transaction commit, but it also can be
    triggered explicitly by a command.

    :param db: The `flask_sqlalchemy.SQLAlchemy` instance

    For example::

        from flask_sqlalchemy import SQLAlchemy
        from flask_signalbus import SignalBus

        app = Flask(__name__)
        db = SQLAlchemy(app)
        signalbus = SignalBus(db)
        signalbus.flush()

    """

    def __init__(self, db, init_app=True):
        self.db = db
        self.signal_session = self.db.create_scoped_session({'expire_on_commit': False})
        self.logger = logging.getLogger(__name__)
        self._autoflush = True
        retry = retry_on_deadlock(self.signal_session, retries=11, max_wait=1.0)
        self._flush_signals_with_retry = retry(self._flush_signals)
        self._flushmany_signals_with_retry = retry(self._flushmany_signals)
        event.listen(self.db.session, 'transient_to_pending', self._transient_to_pending_handler)
        event.listen(self.db.session, 'after_commit', self._safe_after_commit_handler)
        event.listen(self.db.session, 'after_rollback', self._after_rollback_handler)
        event.listen(mapper, 'after_configured', _setup_schema(db.Model, self.db.session))
        if init_app:
            if db.app is None:
                raise RuntimeError(
                    'No application found. The SQLAlchemy instance passed to'
                    ' SignalBus() should be constructed with an application.'
                )
            self._init_app(db.app)

    @property
    def autoflush(self):
        """Setting this property to `False` instructs the `SignalBus` instance
        to not automatically flush pending signals after each
        transaction commit. Setting it back to `True` restores the
        default behavior.

        """

        return self._autoflush

    @autoflush.setter
    def autoflush(self, value):
        self._autoflush = bool(value)

    def get_signal_models(self):
        """Return all signal types in a list.

        :rtype: list(`signal-model`)

        """

        base = self.db.Model
        return [
            cls for cls in _get_class_registry(base).values() if (
                isinstance(cls, type)
                and issubclass(cls, base)
                and hasattr(cls, 'send_signalbus_message')
            )
        ]

    def flush(self, models=None, wait=3.0):
        """Send pending signals over the message bus.

        This method assumes that auto-flushing is enabled for the
        given signal types, and therefore the number of pending
        signals is not too big. Having multiple processes that run
        this method in parallel is generally *not a good idea*.

        :param models: If passed, flushes only signals of the specified types.
        :type models: list(`signal-model`) or `None`
        :param float wait: The number of seconds the method will wait
            after obtaining the list of pending signals, to allow
            auto-flushing senders to complete
        :return: The total number of signals that have been sent

        """

        sent_count = 0
        try:
            models_to_flush = self.get_signal_models() if models is None else models
            pks_by_model = {}
            for model in models_to_flush:
                _raise_error_if_not_signal_model(model)
                pks_by_model[model] = self._list_signal_pks(model)
            self.signal_session.rollback()
            time.sleep(wait)
            for model in models_to_flush:
                self.logger.info('Flushing %s.', model.__name__)
                sent_count += self._flush_signals_with_retry(model, pk_values_set=set(pks_by_model[model]))
        finally:
            self.signal_session.remove()
        return sent_count

    def flushmany(self, models=None):
        """Send a potentially huge number of pending signals over the message bus.

        This method assumes that the number of pending signals might
        be huge. Using `SignalBus.flushmany` when auto-flushing is
        enabled for the given signal types is not recommended, because
        it may result in multiple delivery of messages.

        `SignalBus.flushmany` can be very useful when recovering from
        long periods of disconnectedness from the message bus, or when
        auto-flushing is disabled. If your database (and its
        SQLAlchemy dialect) supports ``FOR UPDATE SKIP LOCKED``,
        multiple processes will be able to run this method in
        parallel, without stepping on each others' toes.

        :param models: If passed, flushes only signals of the specified types.
        :type models: list(`signal-model`) or `None`
        :return: The total number of signals that have been sent

        """

        return self._flush_models(flush_fn=self._flushmany_signals_with_retry, models=models)

    def flushordered(self, models=None):
        """Send all pending messages in predictable order.

        The order is defined by the ``signalbus_order_by`` attribute
        of the model class. When auto-flushing is disabled for the
        given signal types, this method guarantes that messages will
        be sent in the correct order. Having multiple processes that
        run this method in parallel is generally *not a good idea*.

        :param models: If passed, flushes only signals of the specified types.
        :type models: list(`signal-model`) or `None`
        :return: The total number of signals that have been sent

        """

        return self._flush_models(flush_fn=self._flushordered_signals, models=models)

    def _init_app(self, app):
        from . import signalbus_cli

        if not hasattr(app, 'extensions'):
            app.extensions = {}
        if app.extensions.get('signalbus') not in [None, self]:
            raise RuntimeError('Can not attach more than one SignalBus to one application.')
        app.extensions['signalbus'] = self
        app.cli.add_command(signalbus_cli.signalbus)

        @app.teardown_appcontext
        def shutdown_signal_session(response_or_exc):
            self.signal_session.remove()
            return response_or_exc

    def _transient_to_pending_handler(self, session, instance):
        model = type(instance)
        if hasattr(model, 'send_signalbus_message') and getattr(model, 'signalbus_autoflush', True):
            signals_to_flush = session.info.setdefault(_SIGNALS_TO_FLUSH_SESSION_INFO_KEY, set())
            signals_to_flush.add(instance)

    def _after_commit_handler(self, session):
        signals_to_flush = session.info.pop(_SIGNALS_TO_FLUSH_SESSION_INFO_KEY, set())
        if self.autoflush and signals_to_flush:
            signals = [self.signal_session.merge(s, load=False) for s in signals_to_flush]
            for signal in signals:
                try:
                    signal.send_signalbus_message()
                except Exception:
                    self.logger.exception('Caught error while sending %s.', signal)
                    self.signal_session.rollback()
                    return
                self.signal_session.delete(signal)
            self.signal_session.commit()
            self.signal_session.expire_all()

    def _after_rollback_handler(self, session):
        session.info.pop(_SIGNALS_TO_FLUSH_SESSION_INFO_KEY, None)

    def _safe_after_commit_handler(self, session):
        try:
            return self._after_commit_handler(session)
        except DBAPIError as e:
            if get_db_error_code(e.orig) not in DEADLOCK_ERROR_CODES:
                self.logger.exception('Caught database error during autoflush.')
            self.signal_session.rollback()

    def _compose_signal_query(self, model, pk_only=False, ordered=False, max_count=None):
        m = inspect(model)
        pk_attrs = [m.get_property_by_column(c).class_attribute for c in m.primary_key]
        if pk_only:
            query = self.signal_session.query(*pk_attrs)
        else:
            query = self.signal_session.query(model)
        if ordered:
            order_by_columns = getattr(model, 'signalbus_order_by', ())
            if order_by_columns:
                query = query.order_by(*order_by_columns)
        if max_count is not None:
            query = query.limit(max_count)
        return query, pk_attrs

    def _lock_signal_instances(self, model, pk_values_list, ordered=False):
        query, pk_attrs = self._compose_signal_query(model, ordered=ordered)
        clause = or_(*[
            and_(*[attr == value for attr, value in zip(pk_attrs, pk_values)])
            for pk_values in pk_values_list
        ])
        return query.filter(clause).with_for_update().all()

    def _list_signal_pks(self, model, ordered=False, max_count=None):
        query, _ = self._compose_signal_query(model, pk_only=True, ordered=ordered, max_count=max_count)
        return query.all()

    def _get_signal_burst_count(self, model):
        burst_count = int(getattr(model, 'signalbus_burst_count', 1))
        assert burst_count > 0, '"signalbus_burst_count" must be positive'
        return burst_count

    def _send_and_delete_signal_instances(self, model, instances):
        n = len(instances)
        if n > 1 and hasattr(model, 'send_signalbus_messages'):
            model.send_signalbus_messages(instances)
        else:
            for instance in instances:
                instance.send_signalbus_message()
        for instance in instances:
            self.signal_session.delete(instance)
        return n

    def _flush_models(self, flush_fn, models):
        sent_count = 0
        try:
            models_to_flush = self.get_signal_models() if models is None else models
            for model in models_to_flush:
                _raise_error_if_not_signal_model(model)
                sent_count += flush_fn(model)
        finally:
            self.signal_session.remove()
        return sent_count

    def _flushordered_signals(self, model):
        self.logger.info('Flushing %s in "flushordered" mode.', model.__name__)
        sent_count = 0
        while True:
            n = self._flush_signals(model, ordered=True)
            sent_count += n
            if n < _FLUSH_SIGNALS_LIMIT:
                return sent_count

    def _flushmany_signals(self, model):
        self.logger.info('Flushing %s in "flushmany" mode.', model.__name__)
        sent_count = 0
        burst_count = self._get_signal_burst_count(model)
        query, _ = self._compose_signal_query(model, max_count=burst_count)
        query = query.with_for_update(skip_locked=True)
        while True:
            signals = query.all()
            sent_count += self._send_and_delete_signal_instances(model, signals)
            self.signal_session.commit()
            self.signal_session.expire_all()
            if len(signals) < burst_count:
                break
        return sent_count

    def _flush_signals(self, model, pk_values_set=None, ordered=False):
        sent_count = 0
        burst_count = self._get_signal_burst_count(model)
        signal_pks = self._list_signal_pks(model, ordered=ordered, max_count=_FLUSH_SIGNALS_LIMIT)
        self.signal_session.rollback()
        if pk_values_set is not None:
            signal_pks = [pk for pk in signal_pks if pk in pk_values_set]
        for pk_values_chunk in _chunks(signal_pks, size=burst_count):
            signals = self._lock_signal_instances(model, pk_values_chunk, ordered=ordered)
            sent_count += self._send_and_delete_signal_instances(model, signals)
            self.signal_session.commit()
            self.signal_session.expire_all()
        return sent_count


def _setup_schema(Base, session):
    """Create a function which adds `__marshmallow__` attribute to all signal models."""

    def create_schema_class(m):
        class Meta(object):
            model = m
            include_relationships = True
            load_instance = True

        if not hasattr(m, 'send_signalbus_message'):
            # Signal models should not use the SQLAlchemy session.
            Meta.sqla_session = session

        schema_class_name = '%sSchema' % m.__name__
        return type(schema_class_name, (SQLAlchemyAutoSchema,), {'Meta': Meta})

    def setup_schema_fn():
        for model in _get_class_registry(Base).values():
            if hasattr(model, '__tablename__'):
                if model.__name__.endswith("Schema"):
                    raise ModelConversionError(
                        'Unexpected model name: "{}". '
                        'For safety, _setup_schema() can not be used when a '
                        'model class ends with "Schema".'.format(model.__name__)
                    )
                schema_class = getattr(model, '__marshmallow__', None)
                if schema_class is None:
                    schema_class = model.__marshmallow__ = create_schema_class(model)
                schema_class_instance = getattr(model, '__marshmallow_schema__', None)
                if schema_class_instance is None and hasattr(model, 'send_signalbus_message'):
                    setattr(model, '__marshmallow_schema__', schema_class())

    return setup_schema_fn



import pytest
import time
from flask_signalbus import rabbitmq
from flask_signalbus.rabbitmq.publisher import DeliverySet


@pytest.fixture(params=['direct', 'init_app'])
def publisher(app, request):
    if request.param == 'direct':
        p = rabbitmq.Publisher(app)
    elif request.param == 'init_app':
        p = rabbitmq.Publisher()
        p.init_app(app)
    return p


def test_delivery_set():
    s = DeliverySet(20, 10)  # from 20 to 29
    assert(not s.all_confirmed)
    assert(s.confirm(23) is True)
    assert(s.confirm(23) is False)
    assert(not s.all_confirmed)
    assert(s.confirm(10) is False)
    assert(not s.all_confirmed)
    assert(s.confirm(28, multiple=True) is True)
    assert(s.confirm(24, multiple=False) is False)
    assert(not s.all_confirmed)
    assert(s.confirm(29) is True)
    assert(s.all_confirmed)


def test_publisher(publisher):
    message = rabbitmq.Message(
        body='test message',
        properties=rabbitmq.MessageProperties(),
        exchange='',
        routing_key='test',
    )
    publisher.publish_messages([])
    publisher.publish_messages([message])
    publisher.publish_messages([message, message])
    time.sleep(20)
    publisher.publish_messages([message], timeout=60)



from flask_signalbus.signalbus import (  # noqa: F401
    SignalBus,
    SignalBusMixin,
)

from flask_signalbus.atomic import (  # noqa: F401
    AtomicProceduresMixin,
)

from flask_signalbus.utils import DBSerializationError  # noqa: F401



"""
Adds to Flask-SQLAlchemy simple but powerful utilities for
creating consistent and correct database APIs.
"""

from functools import wraps
from contextlib import contextmanager
from sqlalchemy.sql.expression import and_
from sqlalchemy.inspection import inspect
from sqlalchemy.exc import IntegrityError
from flask_signalbus.utils import DBSerializationError, retry_on_deadlock

__all__ = ['AtomicProceduresMixin']


_ATOMIC_FLAG_SESSION_INFO_KEY = 'flask_signalbus__atomic_flag'


class _ModelUtilitiesMixin(object):
    @classmethod
    def get_instance(cls, instance_or_pk, *options):
        """Return a model instance in ``db.session`` or `None`.

        :param instance_or_pk: An instance of this model class, or a
          primary key. A composite primary key can be passed as a
          tuple.
        :param options: Arguments to be passed to
          :meth:`~sqlalchemy.orm.query.Query.options`.

        Example::

          @db.atomic
          def increase_account_balance(account, amount):
              # Here `Account` is a subclass of `db.Model`.
              account = Account.get_instance(account)
              account.balance += amount

          # Now `increase_account_balance` can be
          # called with an account instance:
          increase_account_balance(my_account, 100.00)

          # or with an account primary key (1234):
          increase_account_balance(1234, 100.00)

        """

        if isinstance(instance_or_pk, cls):
            if instance_or_pk in cls._flask_signalbus_sa.session:
                return instance_or_pk
            instance_or_pk = inspect(cls).primary_key_from_instance(instance_or_pk)
        return cls.query.options(*options).get(instance_or_pk)

    @classmethod
    def lock_instance(cls, instance_or_pk, *options, **kw):
        """Return a locked model instance in ``db.session`` or `None`.

        :param instance_or_pk: An instance of this model class, or a
          primary key. A composite primary key can be passed as a
          tuple.
        :param options: Arguments to be passed to
          :meth:`~sqlalchemy.orm.query.Query.options`.
        :param kw: Arguments to be passed to
          :meth:`~sqlalchemy.orm.query.Query.with_for_update`.

        """

        mapper = inspect(cls)
        pk_attrs = [mapper.get_property_by_column(c).class_attribute for c in mapper.primary_key]
        pk_values = cls.get_pk_values(instance_or_pk)
        clause = and_(*[attr == value for attr, value in zip(pk_attrs, pk_values)])
        return cls.query.options(*options).filter(clause).with_for_update(**kw).one_or_none()

    @classmethod
    def get_pk_values(cls, instance_or_pk):
        """Return a primary key as a tuple.

        :param instance_or_pk: An instance of this model class, or a
          primary key. A composite primary key can be passed as a
          tuple.

        """

        if isinstance(instance_or_pk, cls):
            cls._flask_signalbus_sa.session.flush()
            instance_or_pk = inspect(cls).primary_key_from_instance(instance_or_pk)
        return instance_or_pk if isinstance(instance_or_pk, tuple) else (instance_or_pk,)


class AtomicProceduresMixin(object):
    """A **mixin class** that adds utility functions to
    :class:`flask_sqlalchemy.SQLAlchemy` and the declarative base.

    For example::

      from flask_sqlalchemy import SQLAlchemy
      from flask_signalbus import AtomicProceduresMixin

      class CustomSQLAlchemy(AtomicProceduresMixin, SQLAlchemy):
          pass

      db = CustomSQLAlchemy()

      # Now `AtomicProceduresMixin` method are available in `db`.

    Note that when subclassing, `AtomicProceduresMixin` should always
    come before :class:`flask_sqlalchemy.SQLAlchemy`. Adding
    `AtomicProceduresMixin` has several useful results:

    1. `AtomicProceduresMixin` method will be available in ``db``.

    2. The classmethods from
       :class:`~flask_signalbus.atomic._ModelUtilitiesMixin` will be
       available in the declarative base class (``db.Model``), and
       therefore in every model class.

    3. Database isolation level will be set to ``REPEATABLE_READ``.

    """

    def apply_driver_hacks(self, app, info, options):
        if info.drivername != 'sqlite' and "isolation_level" not in options:
            options["isolation_level"] = "REPEATABLE_READ"
        return super(AtomicProceduresMixin, self).apply_driver_hacks(app, info, options)

    def make_declarative_base(self, model, *args, **kwargs):
        class model(_ModelUtilitiesMixin, model):
            pass
        declarative_base = super(AtomicProceduresMixin, self).make_declarative_base(model, *args, **kwargs)
        declarative_base._flask_signalbus_sa = self
        return declarative_base

    def atomic(self, func):
        """A decorator that wraps a function in an atomic block.

        Example::

          db = CustomSQLAlchemy()

          @db.atomic
          def f():
              write_to_db('a message')
              return 'OK'

          assert f() == 'OK'

        This code defines the function ``f``, which is wrapped in an
        atomic block. Wrapping a function in an atomic block gives
        several guarantees:

        1. The database transaction will be automatically committed if
           the function returns normally, and automatically rolled
           back if the function raises unhandled exception.

        2. When the transaction is committed, all objects in
           ``db.session`` will be expunged. This means that no lazy
           loading will be performed on them.

        3. If a transaction serialization error occurs during the
           execution of the function, the function will be
           re-executed. (It might be re-executed several times.)

        Atomic blocks can be nested, but in this case the outermost
        block takes full control of transaction's life-cycle, and
        inner blocks do nothing.

        """

        @wraps(func)
        def wrapper(*args, **kwargs):
            session = self.session
            session_info = session.info
            if session_info.get(_ATOMIC_FLAG_SESSION_INFO_KEY):
                return func(*args, **kwargs)

            @retry_on_deadlock(session)
            def f(*args, **kwargs):
                r = func(*args, **kwargs)
                session.flush()
                return r

            session_info[_ATOMIC_FLAG_SESSION_INFO_KEY] = True
            try:
                result = f(*args, **kwargs)
                session.expunge_all()
                session.commit()
                return result
            except Exception:
                session.rollback()
                raise
            finally:
                session_info[_ATOMIC_FLAG_SESSION_INFO_KEY] = False

        return wrapper

    def execute_atomic(self, func):
        """A decorator that executes a function in an atomic block (see :meth:`atomic`).

        Example::

          db = CustomSQLAlchemy()

          @db.execute_atomic
          def result():
              write_to_db('a message')
              return 'OK'

          assert result == 'OK'

        This code defines *and executes* the function ``result`` in an
        atomic block. At the end, the name ``result`` holds the value
        returned from the function.
        """

        return self.atomic(func)()

    @contextmanager
    def retry_on_integrity_error(self):
        """Re-raise :class:`~sqlalchemy.exc.IntegrityError` as `DBSerializationError`.

        This is mainly useful to handle race conditions in atomic
        blocks. For example, even if prior to a database INSERT we
        have verified that there is no existing row with the given
        primary key, we still may get an
        :class:`~sqlalchemy.exc.IntegrityError` if another transaction
        inserted a row with this primary key in the meantime. But if
        we do (within an atomic block)::

          with db.retry_on_integrity_error():
              db.session.add(instance)

        then if the before-mentioned race condition occurs,
        `DBSerializationError` will be raised instead of
        :class:`~sqlalchemy.exc.IntegrityError`, so that the
        transaction will be retried (by the atomic block), and the
        second time our prior-to-INSERT check will correctly detect a
        primary key collision.

        Note: :meth:`retry_on_integrity_error` triggers a session
        flush.

        """

        session = self.session
        assert session.info.get(_ATOMIC_FLAG_SESSION_INFO_KEY), \
            'Calls to "retry_on_integrity_error" must be wrapped in atomic block.'
        session.flush()
        try:
            yield
            session.flush()
        except IntegrityError:
            raise DBSerializationError



"""
Flask-SignalBus
---------------

Adds to Flask-SQLAlchemy the capability to *atomically* send messages
(signals) over a message bus.

The processing of each message involves three steps:

  1. The message is recorded in the SQL database as a row in a table.

  2. The message is sent over the message bus (RabbitMQ for example).

  3. Message's corresponding table row is deleted.

Normally, the sending of the recorded messages (steps 2 and 3) is done
automatically after each transaction commit, but when needed, it can
also be triggered explicitly with a method call, or through the Flask
CLI.
"""

import sys
from setuptools import setup

needs_pytest = {'pytest', 'test', 'ptr'}.intersection(sys.argv)
pytest_runner = ['pytest-runner'] if needs_pytest else []


setup(
    name='Flask-SignalBus',
    version='0.5.11',
    url='https://github.com/epandurski/flask_signalbus',
    license='MIT',
    author='Evgeni Pandurski',
    author_email='epandurski@gmail.com',
    description='A Flask-SQLAlchemy extension for atomically sending messages (signals) over a message bus',
    long_description=__doc__,
    packages=['flask_signalbus', 'flask_signalbus.rabbitmq'],
    zip_safe=True,
    platforms='any',
    setup_requires=pytest_runner,
    install_requires=[
        'Flask-SQLAlchemy>=1.0',
        'marshmallow-sqlalchemy>=0.22.0',
    ],
    tests_require=[
        'pytest~=6.2',
        'pytest-cov~=2.7',
        'mock~=2.0',
        'pika~=1.3',
    ],
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ]
)



import sys
import time
import logging
import click
from flask.cli import with_appcontext
from flask import current_app
from . import SignalBus
from .utils import report_signal_count


def _get_models_to_flush(signalbus, signal_names, exclude):
    signal_names = set(signal_names)
    exclude = set(exclude)
    models_to_flush = signalbus.get_signal_models()
    if signal_names and exclude:
        logger = logging.getLogger(__name__)
        logger.warning('Specified both SIGNAL_NAMES and exclude option.')
    if signal_names:
        wrong_signal_names = signal_names - {m.__name__ for m in models_to_flush}
        models_to_flush = [m for m in models_to_flush if m.__name__ in signal_names]
    else:
        wrong_signal_names = exclude - {m.__name__ for m in models_to_flush}
    for name in wrong_signal_names:
        logger = logging.getLogger(__name__)
        logger.warning('A signal with name "%s" does not exist.', name)
    return [m for m in models_to_flush if m.__name__ not in exclude]


@click.group()
def signalbus():
    """Perform SignalBus operations."""


@signalbus.command()
@with_appcontext
@click.option('-e', '--exclude', multiple=True, help='Do not flush signals with the specified name.')
@click.option('-w', '--wait', type=float, help='The number of seconds "flush" will wait after'
              ' obtaining the list of pending signals, to allow concurrent senders to complete.'
              ' The default is %s seconds.' % SignalBus.flush.__defaults__[1])
@click.option('-r', '--repeat', type=float, help='Flush every FLOAT seconds.')
@click.argument('signal_names', nargs=-1)
def flush(signal_names, exclude, wait, repeat):
    """Send pending signals over the message bus.

    If a list of SIGNAL_NAMES is specified, flushes only those
    signals. If no SIGNAL_NAMES are specified, flushes all signals.

    This method assumes that auto-flushing is enabled for the given
    signal types, and therefore the number of pending signals is not
    too big. Having multiple processes that run this method in
    parallel is generally not a good idea.

    """

    signalbus = current_app.extensions['signalbus']
    models_to_flush = _get_models_to_flush(signalbus, signal_names, exclude)
    logger = logging.getLogger(__name__)
    logger.info('Started flushing %s.', ', '.join(m.__name__ for m in models_to_flush))

    while True:
        started_at = time.time()
        try:
            if wait is not None:
                signal_count = signalbus.flush(models_to_flush, wait=max(0.0, wait))
            else:
                signal_count = signalbus.flush(models_to_flush)
        except Exception:
            logger.exception('Caught error while sending pending signals.')
            sys.exit(1)

        report_signal_count(signal_count)

        if repeat is None:
            break
        else:
            time.sleep(max(0.0, repeat + started_at - time.time()))


@signalbus.command()
@with_appcontext
@click.option('-e', '--exclude', multiple=True, help='Do not flush signals with the specified name.')
@click.option('-r', '--repeat', type=float, help='Flush every FLOAT seconds.')
@click.argument('signal_names', nargs=-1)
def flushmany(signal_names, exclude, repeat):
    """Send a potentially huge number of pending signals over the message bus.

    If a list of SIGNAL_NAMES is specified, flushes only those
    signals. If no SIGNAL_NAMES are specified, flushes all signals.

    This command assumes that the number of pending signals might be
    huge. Using "flushmany" when auto-flushing is enabled for the
    given signal types is not recommended, because it may result in
    multiple delivery of messages.

    "flushmany" can be very useful when recovering from long periods
    of disconnectedness from the message bus, or when auto-flushing is
    disabled. If your database (and its SQLAlchemy dialect) supports
    "FOR UPDATE SKIP LOCKED", multiple processes will be able to run
    this command in parallel, without stepping on each others' toes.

    """

    signalbus = current_app.extensions['signalbus']
    models_to_flush = _get_models_to_flush(signalbus, signal_names, exclude)
    logger = logging.getLogger(__name__)
    logger.info('Started flushing %s.', ', '.join(m.__name__ for m in models_to_flush))

    while True:
        started_at = time.time()
        try:
            signal_count = signalbus.flushmany(models_to_flush)
        except Exception:
            logger.exception('Caught error while sending pending signals.')
            sys.exit(1)

        report_signal_count(signal_count)

        if repeat is None:
            break
        else:
            time.sleep(max(0.0, repeat + started_at - time.time()))


@signalbus.command()
@with_appcontext
@click.option('-e', '--exclude', multiple=True, help='Do not flush signals with the specified name.')
@click.option('-r', '--repeat', type=float, help='Flush every FLOAT seconds.')
@click.argument('signal_names', nargs=-1)
def flushordered(signal_names, exclude, repeat):
    """Send all pending messages in predictable order.

    If a list of SIGNAL_NAMES is specified, flushes only those
    signals. If no SIGNAL_NAMES are specified, flushes all signals.

    The order is defined by the "signalbus_order_by" attribute of the
    model class. When auto-flushing is disabled for the given signal
    types, this method guarantes that messages will be sent in the
    correct order. Having multiple processes that run this method in
    parallel is generally not a good idea.

    """

    signalbus = current_app.extensions['signalbus']
    models_to_flush = _get_models_to_flush(signalbus, signal_names, exclude)
    logger = logging.getLogger(__name__)
    logger.info('Started flushing %s.', ', '.join(m.__name__ for m in models_to_flush))

    while True:
        started_at = time.time()
        try:
            signal_count = signalbus.flushordered(models_to_flush)
        except Exception:
            logger.exception('Caught error while sending pending signals.')
            sys.exit(1)

        report_signal_count(signal_count)

        if repeat is None:
            break
        else:
            time.sleep(max(0.0, repeat + started_at - time.time()))


@signalbus.command()
@with_appcontext
def signals():
    """Show all signal types."""

    signalbus = current_app.extensions['signalbus']
    for signal_model in signalbus.get_signal_models():
        click.echo(signal_model.__name__)


@signalbus.command()
@with_appcontext
def pending():
    """Show the number of pending signals by signal type."""

    signalbus = current_app.extensions['signalbus']
    pending = []
    total_pending = 0
    for signal_model in signalbus.get_signal_models():
        count = signal_model.query.count()
        if count > 0:
            pending.append((count, signal_model.__name__))
        total_pending += count
    if pending:
        pending.sort()
        max_chars = len(str(pending[-1][0]))
        for n, signal_name in pending:
            click.echo('{} of type "{}"'.format(str(n).rjust(max_chars), signal_name))
    click.echo(25 * '-')
    click.echo('Total pending: {} '.format(total_pending))
